# Package: PENTAtrainer2
# Version: 1.1.2
# Filename: synthesize.praat
#
# Author: Santitham Prom-on, Yi Xu
# Last updated: 20 December 2016
#
# Objective: synthesize the F0 contours based on categorical parameters 
# specified in parameter.txt

directory$ = chooseDirectory$ ("Choose a directory to synthesize")
if directory$ = ""
	#[EXIT_00] Choose to not Continue
	exit Working directory must be selected
else
	#Check validity of the directory
	Create Strings as file list... FileList 'directory$'/*.wav
	numberOfFiles = Get number of strings
	if !numberOfFiles
		select Strings FileList
		Remove
		Create Strings as file list... FileList 'directory$'/*.WAV
		numberOfFiles = Get number of strings
	endif
	
	#[EXIT_S01] No sound files in the directory	
	if numberOfFiles = 0
		exit Error Code A01: No sound files in the directory
	else
		Write to raw text file... 'directory$'/FileList.txt
		more_file = 1
	endif
endif

### Error Checking ###

#[EXIT_S02] Cannot find the parameter file (parameters.txt)
if not fileReadable("'directory$'/parameters.txt")
	exit Error Code L03: Cannot find the parameter file (parameters.txt)
endif

Create Strings as file list... aList 'directory$'/*.annotation
n_a = Get number of strings
if n_a = 0
	#[EXIT_S03] Cannot find annotation files
	exit Error Code L03: Cannot find annotation files	
elsif n_a < numberOfFiles
	#[EXIT_S04] Number of annotation files must equal number of sound files
	exit Error Code S04: Number of annotation files must equal number of sound files
endif
for i from 1 to n_a
	select Strings aList
	fileName$ = Get string... i
	name$ = fileName$ - ".annotation"
	nocheck Read from file... 'directory$'/'fileName$'
	fullName$ = selected$()
	name1$ = extractLine$(fullName$," ")
	if (name$ = name1$)
		Down to Table... no 4 yes no
		Save as tab-separated file... 'directory$'/'name$'.intervals
	
		select TextGrid 'name$'
		plus Table 'name$'
		Remove
	else
		#[EXIT_S05] Wrong format (annotation)
		exit Error Code S05: Wrong format ('name$'.annotation)
	endif
endfor
select Strings aList
Remove


Create Strings as file list... iList 'directory$'/*.intervals
n_i = Get number of strings
if n_i = 0
	#[EXIT_S06] Cannot find intervals files
	exit Error Code S06: Cannot find intervals files	
elsif n_i < numberOfFiles
	#[EXIT_S07] Number of intervals files must equal number of sound files
	exit Error Code S07: Number of intervals files must equal number of sound files
endif
for i from 1 to n_i
	select Strings iList
	fileName$ = Get string... i
	name$ = fileName$ - ".intervals"
	nocheck Read from file... 'directory$'/'fileName$'
	fullName$ = selected$()
	name1$ = extractLine$(fullName$," ")
	if (name$ = name1$)
		Remove
	else
		#[EXIT_S08] Wrong format (intervals)
		exit Error Code S08: Wrong format ('name$'.intervals)
	endif
endfor
select Strings iList
Remove


Create Strings as file list... fList 'directory$'/*.actutimenormf0
n_f = Get number of strings
if n_f = 0
	#[EXIT_S09] Cannot find actutimenormf0 files
	exit Error Code S09: Cannot find actutimenormf0 files	
elsif n_f < numberOfFiles
	#[EXIT_S10] Number of actutimenormf0 files must equal number of sound files
	exit Error Code S10: Number of actutimenormf0 files must equal number of sound files
endif
for i from 1 to n_f
	select Strings fList
	fileName$ = Get string... i
	name$ = fileName$ - ".actutimenormf0"
	nocheck Read from file... 'directory$'/'fileName$'
	fullName$ = selected$()
	name1$ = extractLine$(fullName$," ")
	if (name$ = name1$)
		Remove
	else
		#[EXIT_S11] Wrong format (actutimenormf0)
		exit Error Code S11: Wrong format ('name$'.actutimenormf0)
	endif
endfor
select Strings fList
Remove

Create Strings as file list... sfList 'directory$'/*.semitonef0
n_sf = Get number of strings
if n_sf = 0
	#[EXIT_S12] Cannot find actutimenormf0 files
	exit Error Code S12: Cannot find actutimenormf0 files	
elsif n_sf < numberOfFiles
	#[EXIT_S13] Number of actutimenormf0 files must equal number of sound files
	exit Error Code S13: Number of actutimenormf0 files must equal number of sound files
endif
for i from 1 to n_sf
	select Strings sfList
	fileName$ = Get string... i
	name$ = fileName$ - ".semitonef0"
	nocheck Read from file... 'directory$'/'fileName$'
	fullName$ = selected$()
	name1$ = extractLine$(fullName$," ")
	if (name$ = name1$)
		Remove
	else
		#[EXIT_S14] Wrong format (actutimenormf0)
		exit Error Code S14: Wrong format ('name$'.actutimenormf0)
	endif
endfor
select Strings sfList
Remove

### End of Error Checking ###

beginPause("Synthesize - Options")
	integer("Input File No",1)
	choice("Task:",1)
		option("Interactive view")
		option("Synthesize all sounds without pause")
	boolean("Inspect Manipulation",0)
	boolean("Save synthesized sound",0)
clicked = endPause("Cancel","Start",2,1)

#[EXIT_00] Choose to not Continue
if clicked = 1
	exit
endif

if save_synthesized_sound = 1
	outputdir$ = chooseDirectory$("Choose the output directory")
	if outputdir$ = ""
		exit Output directory must be selected to write the synthesized sound
	endif
endif

max_strength = 100

Create Table with column names... accuracy numberOfFiles+1 Filename RMSE Correlation

mean_rmse = 0
mean_corr = 0
for current_file from input_File_No to numberOfFiles
	select Strings FileList
	fileName$ = Get string... current_file
	name$ = fileName$ - ".wav" - ".WAV"
	system java -jar Synthesize.jar "'directory$'" 'name$' parameters.txt

	Read from file... 'directory$'/'name$'.synthesizedf0
	nrow = Get number of rows
	start_time = Get value... 1 Time
	end_time = Get value... nrow Time
	Create PitchTier... 'name$'_syn start_time end_time
	for i from 1 to nrow
		select Table 'name$'
		t = Get value... i Time
		f0 = Get value... i F0
		select PitchTier 'name$'_syn
		Add point... 't' 'f0'
	endfor

	Read from file... 'directory$'/'name$'.syn_acc
	Rename... 'name$'_acc
	rmse = Get value... 1 RMSE
	corr = Get value... 1 CORR
	mean_rmse += rmse
	mean_corr += corr

	select Table accuracy
	Set string value... current_file Filename 'name$'
	Set numeric value... current_file RMSE 'rmse'
	Set numeric value... current_file Correlation 'corr'

	Read from file... 'directory$'/'fileName$'
	To Manipulation... 0.01 75 500
	plus PitchTier 'name$'_syn
	Replace pitch tier

	if save_synthesized_sound
		select Manipulation 'name$'
		Get resynthesis (overlap-add)
		Save as WAV file... 'outputdir$'/'name$'_syn.wav
		Remove
	endif

	select Table 'name$'
	plus Table 'name$'_acc
	Remove

	Read from file... 'directory$'/'name$'.PitchTier
	Read from file... 'directory$'/'name$'.annotation
	Read from file... 'directory$'/'name$'.targets
	
	if task = 1
		select Sound 'name$'
		tInitial = Get start time
		tFinal = Get end time
		call plot_target tInitial tFinal
	endif
	select Sound 'name$'
	plus PitchTier 'name$'
	plus PitchTier 'name$'_syn
	if inspect_Manipulation = 0
		plus Manipulation 'name$'
	endif
	plus TextGrid 'name$'
	plus Table 'name$'
	Remove
endfor
mean_rmse /= numberOfFiles
mean_corr /= numberOfFiles
if task = 2
	select Table accuracy
	Set string value... numberOfFiles+1 Filename Mean
	Set numeric value... numberOfFiles+1 RMSE 'mean_rmse'
	Set numeric value... numberOfFiles+1 Correlation 'mean_corr'
	Save as tab-separated file... 'directory$'/accuracy_synthesis.txt
	plus Strings FileList
	Remove
	printline Synthesis completed!
	printline mean RMSE = 'mean_rmse'		mean Correlation = 'mean_corr'
endif

procedure plot_target t1 t2
	demoWindowTitle ("PENTAtrainer 2")
	demo Erase all
	demo Paint rectangle... White 0 100 -100 100
	demo Blue
	demo Times
	demo Line width... 2
	demo 12
	demo Select outer viewport... 0 100 50 100
	select PitchTier 'name$'
	To Pitch... 0.02 60 600
	Formula... 12 * ln (self) / ln(2); semitone
	f0_floor = Get minimum... 0 0 Hertz Parabolic
	f0_ceiling = Get maximum... 0 0 Hertz Parabolic
	f0_floor -= 10
	f0_ceiling += 10
	demo Draw... t1 t2 f0_floor f0_ceiling no
	Remove
	demo Line width... 0.5
	demo Draw inner box

	select PitchTier 'name$'_syn
	Copy... 'name$'_synst
	Formula... 12 * ln (self) / ln(2); semitone
	demo Black
	demo Text special... t1-0.04*(t2-t1) Centre 0.5*(f0_floor+f0_ceiling) Bottom Times 12 90 Pitch (st)
	demo Text special... 0.5*(t1+t2) Centre f0_floor-0.125*(f0_ceiling-f0_floor) Top Times 12 0 Time (s)
	demo Red
	demo Draw... t1 t2 f0_floor f0_ceiling no
	Remove

	demo Marks bottom every... 1 0.1 no yes no
	demo Marks bottom every... 1 0.5 yes yes no
	demo Marks left every... 1 5 no yes no
	demo Marks left every... 1 10 yes yes no

	select TextGrid 'name$'
	nintervals = Get number of intervals... 1

	# Remove lines in voiceless regions
	last_end = 32767
	for nn from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 nn
		if not label$ = ""
			start = Get starting point... 1 nn
			end = Get end point... 1 nn
			if (start - last_end) > 0.001
				demo Paint rectangle... white last_end start f0_floor+1 f0_ceiling-1
			endif
			last_end = end
		elsif nn = 1 or nn = nintervals
			start = Get starting point... 1 nn
			end = Get end point... 1 nn
			demo Paint rectangle... white start end-0.001 f0_floor+1 f0_ceiling-1
		endif
	endfor

	# Draw pitch targets and syllable boundaries
	interval_plot = 0
	last_end = 32767
	demo Dashed line
	for nn from 1 to nintervals
		select TextGrid 'name$'
		label$ = Get label of interval... 1 nn
		if not label$ = ""
			interval_plot += 1
			start = Get starting point... 1 nn
			end = Get end point... 1 nn
			duration = end - start
			select Table 'name$'
			m = Get value... 'interval_plot' m
			b = Get value... 'interval_plot' b
			lambda = Get value... 'interval_plot' r
			strength = if lambda > 0 then 3 * lambda / max_strength else 0.001 fi
			y1 = b - m*duration
			y2 = b
			# demo Black
			# demo Text special... start Left f0_ceiling Bottom Times 12 0 'label$'
			demo Green
			demo Line width... strength
			demo Draw line... start y1 end y2
			demo Grey
			demo Line width... 0.25
			demo Draw line... start f0_floor start f0_ceiling
			if (start - last_end) > 0.001
				demo Draw line... last_end f0_floor last_end f0_ceiling
			endif
			last_end = end
		elsif nn = 1 or nn = nintervals
			start = Get starting point... 1 nn
			demo Grey
			demo Line width... 0.25
			demo Draw line... start f0_floor start f0_ceiling
		endif
	endfor

	# Draw functional layers
	demo Navy
	demo Select outer viewport... 0 100 10 49
	demo Axes... t1 t2 0 10
	demo Text special... t2 right 10 bottom Courier 12 0 Communicative Functions
	demo Draw inner box

		# Draw syllable boundaries
		interval_plot = 0
		last_end = 32767
		demo Dashed line
		demo Grey
		for nn from 1 to nintervals
			select TextGrid 'name$'
			label$ = Get label of interval... 1 nn
			if not label$ = ""
				start = Get starting point... 1 nn
				end = Get end point... 1 nn
				demo Line width... 0.25
				demo Draw line... start 0 start 10
				if (start - last_end) > 0.001
					demo Draw line... last_end 0 last_end 10
				endif
				last_end = end
			elsif nn = 1 or nn = nintervals
				start = Get starting point... 1 nn
				demo Line width... 0.25
				demo Draw line... start 0 start 10
			endif
		endfor

		# Draw labels
		select TextGrid 'name$'
		ntiers = Get number of tiers
		top_level = 10
		for tt from 1 to ntiers
			tier_level = top_level - tt
			tier_name$ = Get tier name... tt
			for nn from 1 to nintervals
				select TextGrid 'name$'
				label$ = Get label of interval... tt nn
				if not label$ = ""
					start = Get starting point... tt nn
					end = Get end point... tt nn
					middle = (start+end)/2
					demo Black
					demo Text special... middle centre tier_level half Times 12 0 'label$'
				endif
			endfor
			demo Maroon
			demo Text special... t2 Right tier_level half Courier 12 0 'tier_name$'
		endfor

	# Draw interfaces
	demo Select inner viewport... 0 95 95 98
	demo Axes... 7 95 0 3
	demo Paint rounded rectangle... {0.98,0.98,0.8} 12 18 0 3 2	; zoom in
	demo Paint rounded rectangle... {0.98,0.98,0.8} 19 25 0 3 2	; zoom out
	demo Paint rounded rectangle... {0.98,0.98,0.8} 26 32 0 3 2	; zoom all
	demo Paint rounded rectangle... {0.98,0.98,0.8} 34 38 0 3 2	; shift to start
	demo Paint rounded rectangle... {0.98,0.98,0.8} 39 43 0 3 2	; shift left
	demo Paint rounded rectangle... {0.98,0.98,0.8} 44 48 0 3 2	; shif right
	demo Paint rounded rectangle... {0.98,0.98,0.8} 49 53 0 3 2	; shift to end
	demo Paint rounded rectangle... {0.98,0.98,0.8} 55 63 0 3 2	; play original
	demo Paint rounded rectangle... {0.98,0.98,0.8} 64 74 0 3 2	; play resynthesis
	demo Paint rounded rectangle... {0.98,0.98,0.8} 76 82 0 3 2	; Previous
	demo Paint rounded rectangle... {0.98,0.98,0.8} 83 87 0 3 2	; Next
	demo Paint rounded rectangle... {0.98,0.98,0.8} 89 93 0 3 2	; Exit
	demo Solid line
	demo Line width... 0.5
	demo Black
	demo Draw rounded rectangle... 12 18 0 3 2
	demo Draw rounded rectangle... 19 25 0 3 2
	demo Draw rounded rectangle... 26 32 0 3 2
	demo Draw rounded rectangle... 34 38 0 3 2
	demo Draw rounded rectangle... 39 43 0 3 2
	demo Draw rounded rectangle... 44 48 0 3 2
	demo Draw rounded rectangle... 49 53 0 3 2
	demo Draw rounded rectangle... 55 63 0 3 2
	demo Draw rounded rectangle... 64 74 0 3 2
	demo Draw rounded rectangle... 76 82 0 3 2
	demo Draw rounded rectangle... 83 87 0 3 2
	demo Draw rounded rectangle... 89 93 0 3 2
	demo Text special... 15 centre 1.5 half Times 10 0 Zoom in
	demo Text special... 22 centre 1.5 half Times 10 0 Zoom out
	demo Text special... 29 centre 1.5 half Times 10 0 Zoom all
	demo Text special... 36 centre 1.5 half Helvetica 10 0 |<<
	demo Text special... 41 centre 1.5 half Helvetica 10 0 <<
	demo Text special... 46 centre 1.5 half Helvetica 10 0 >>
	demo Text special... 51 centre 1.5 half Helvetica 10 0 >>|
	demo Text special... 59 centre 1.5 half Times 10 0 Play original
	demo Text special... 69 centre 1.5 half Times 10 0 Play resynthesis
	demo Text special... 79 centre 1.5 half Times 10 0 Previous
	demo Text special... 85 centre 1.5 half Times 10 0 Next
	demo Text special... 91 centre 1.5 half Times 10 0 Exit
	demo Select inner viewport... 0 50 0 10
	demo Axes... 0 50 0 10
	demo Text special... 2 left 9 bottom Courier 12 0 Current sound: 'name$' (No.'current_file')

	demo Select inner viewport... 50 100 0 10
	demo Axes... 0 50 0 10
	demo Black
	demo Text special... 20 left 10 bottom Helvetica 12 0 Legends

	demo Blue
	demo Line width... 2
	demo Draw line... 20 8 22 8
	demo Text special... 23 left 8 half Helvetica 10 0 Original F0 Contour

	demo Red
	demo Line width... 1
	demo Draw line... 20 6 22 6
	demo Text special... 23 left 6 half Helvetica 10 0 Synthesized F0 Contour

	demo Green
	demo Dashed line
	demo Line width... 1
	demo Draw line... 20 4 22 4
	demo Text special... 23 left 4 half Helvetica 10 0 Pitch Targets

	if more_file = 0
		demo Paint rectangle... White 0 100 25 30
		demo Red
		demo Text special... 10 left 57 bottom Courier 18 0 Last sound file...
	elsif more_file < 0
		demo Paint rectangle... White 0 100 25 30
		demo Red
		demo Text special... 10 left 57 bottom Courier 18 0 First sound file...
	endif
	call on_input
endproc

procedure on_input
	demo Select inner viewport... 0 95 95 98
	demo Axes... 7 95 0 3
	loop = 1
	while (loop and demoWaitForInput ( ))
		if demoClicked ( )
			if demoClickedIn (12, 18, 0, 3)			; clicked Zoom in
				new_window_centre = t1 + (t2-t1)/2
				plot_width = 0.5 * (t2-t1)
				t1 = max (new_window_centre - plot_width/2, tInitial)
				t2 = min (new_window_centre + plot_width/2, tFinal)
				plot_width = t2-t1
				call plot_target t1 t2
			elsif demoClickedIn (19, 25, 0, 3)		; clicked Zoom out
				new_window_centre = t1 + (t2-t1)/2
				plot_width = 2 * (t2-t1)
				t1 = max (new_window_centre - plot_width/2, tInitial)
				t2 = min (new_window_centre + plot_width/2, tFinal)
				plot_width = t2-t1
				call plot_target t1 t2
			elsif demoClickedIn (26, 32, 0, 3)		; clicked Zoom all
				new_window_centre = tInitial + (tFinal-tInitial)/2
				plot_width = 100 * (tFinal-tInitial)
				t1 = max (new_window_centre - plot_width/2, tInitial)
				t2 = min (new_window_centre + plot_width/2, tFinal)
				plot_width = t2-t1
				call plot_target t1 t2
			elsif demoClickedIn (34, 38, 0, 3)			; clicked Shift to start
				call plot_target tInitial t1+plot_width
			elsif demoClickedIn (39, 43, 0, 3)		; clicked Shift left
				t1 = min (t1 - plot_width, tFinal + plot_width)
				t1 = max (t1, tInitial)
				t2 = t1 + plot_width
				call plot_target t1 t2
			elsif demoClickedIn (44, 48, 0, 3)		; clicked Shift right
				t1 = min (t1 + plot_width, tFinal - plot_width)
				t1 = max (t1, tInitial)
				t2 = t1 + plot_width
				call plot_target t1 t2
			elsif demoClickedIn (49, 53, 0, 3)		; clicked Shift to end
				call plot_target tFinal-plot_width tFinal
			elsif demoClickedIn (55, 63, 0, 3)		; clicked Play original
				demo Paint rounded rectangle... Yellow 55 63 0 3 2
				demo Black
				demo Draw rounded rectangle... 55 63 0 3 2
				demo Text special... 59 centre 1.5 half Times 10 0 Play original
				select Sound 'name$'
				Extract part... t1 t2 Rectangular 1.0 yes
				Play
				Remove
				demo Paint rounded rectangle... {0.98,0.98,0.8} 55 63 0 3 2
				demo Black
				demo Draw rounded rectangle... 55 63 0 3 2
				demo Text special... 59 centre 1.5 half Times 10 0 Play original
			elsif demoClickedIn (64, 74, 0, 3)		; clicked Play resynthesis
				# if found_interval
					demo Paint rounded rectangle... Yellow 64 74 0 3 2
					demo Black
					demo Draw rounded rectangle... 64 74 0 3 2
					demo Text special... 69 centre 1.5 half Times 10 0 Play resynthesis
			
					select Manipulation 'name$'
					plus PitchTier 'name$'_syn
					Replace pitch tier
			
					select Manipulation 'name$'
					resyn = Get resynthesis (overlap-add)
					Extract part... t1 t2 Rectangular 1.0 yes
					Play
					plus resyn
					Remove
					# plus sound
					# plus manipulation
			
					demo Paint rounded rectangle... {0.98,0.98,0.8} 64 74 0 3 2
					demo Black
					demo Draw rounded rectangle... 64 74 0 3 2
					demo Text special... 69 centre 1.5 half Times 10 0 Play resynthesis
				# else
					# demo Paint rectangle... White 0 100 25 30
					# demo Red
					# demo Text special... 2 left 27 bottom Courier 14 0 No resynthesis...
				# endif
			elsif demoClickedIn (76, 82, 0, 3)		; clicked Previous
				if current_file < 2
					more_file -= 2
					call plot_target t1 t2
				else
					current_file -= 2
					demo Erase all
					loop = 0
					replot = 0
					more_file = 1
				endif
			elsif demoClickedIn (83, 87, 0, 3)		; clicked Next
				if current_file = numberOfFiles
					more_file = 0
					call plot_target t1 t2
				else
					demo Erase all
					loop = 0
					replot = 0
					more_file = 1
				endif				
			elsif demoClickedIn (89, 93, 0, 3)		; clicked Exit
				demo Select inner viewport... 0 100 50 100
				demo Axes... 0 100 0 2
				demo Red
				demo Text special... 30 left 0.5 bottom Courier 18 0 Now close window manually...
				loop = 0
				select Sound 'name$'
				plus PitchTier 'name$'
				plus PitchTier 'name$'_syn
				plus Manipulation 'name$'
				plus TextGrid 'name$'
				plus Table 'name$'
				Remove
	;					goto EXIT
				exit
			endif
		endif
		demo Select inner viewport... 0 95 95 98
		demo Axes... 7 95 0 3
	endwhile
endproc